# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import String
from suanpan.storage import storage
from suanpan.tools import ToolComponent as tc


@tc.param(String(key="action", required=True))
@tc.param(String(key="local"))
@tc.param(String(key="remote", required=True))
@tc.param(String(key="dist"))
def SPStorageTools(context):
    args = context.args

    if args.action == "download":
        storage.download(args.remote, args.local)
    elif args.action == "upload":
        storage.upload(args.remote, args.local)
    elif args.action == "remove":
        storage.remove(args.remote)
    elif args.action == "copy":
        storage.copy(args.remote, args.dist)
    else:
        raise Exception("Unsupport action: {}".format(args.action))


if __name__ == "__main__":
    SPStorageTools()  # pylint: disable=no-value-for-parameter



# coding=utf-8
from __future__ import absolute_import, print_function

import requests

from suanpan import asyncio, debug
from suanpan.utils import term

DEFAULT_MARK_PERCENTAGES = (0.5, 0.66, 0.75, 0.8, 0.9, 0.95, 0.98, 0.99, 1)


roundns = lambda s: round(s * 1000, 3)


def printResponseTime(results, markPercentages=DEFAULT_MARK_PERCENTAGES):
    number = len(results)
    costTimes = sorted(
        ({"total": r[0], "response": r[1].elapsed.total_seconds()} for r in results),
        key=lambda t: t["response"],
    )
    responseTimes = [t["response"] for t in costTimes]
    sumTime = sum(responseTimes)
    avgTime = sumTime / number
    minTime = min(responseTimes)
    maxTime = max(responseTimes)

    print()
    term.table(
        [
            ["Response Time:"],
            ["---"],
            ["Sum", "{}ms".format(roundns(sumTime))],
            ["Average", "{}ms".format(roundns(avgTime))],
            ["Min", "{}ms".format(roundns(minTime))],
            ["Max", "{}ms".format(roundns(maxTime))],
        ]
    )

    print()
    print("Percentage of the requests served within a certain time:")
    print("---")
    term.table(
        [
            [
                "{}%".format(int(mp * 100)),
                "{}ms".format(roundns(costTimes[int(number * mp) - 1]["total"])),
                "{}ms".format(roundns(costTimes[int(number * mp) - 1]["response"])),
            ]
            for mp in markPercentages
        ],
        headers=["", "Total", "Response"],
    )

    return results


def test(
    func,
    number,
    concurrency=1,
    thread=True,
    args=None,
    kwargs=None,
    title="ABTest",
    funcName=None,
    markPercentages=DEFAULT_MARK_PERCENTAGES,
):
    args = args or []
    kwargs = kwargs or {}
    testFunc = lambda x: debug.costCall(func, *args, **kwargs)
    funcName = funcName or func.__name__
    func.__name__ = funcName
    totalTime, results = debug.costCall(
        asyncio.map,
        testFunc,
        range(number),
        workers=concurrency,
        thread=thread,
        pbar=title,
    )
    costTimes = sorted(r[0] for r in results)
    sumTime = sum(costTimes)
    avgTime = sumTime / number
    minTime = min(costTimes)
    maxTime = max(costTimes)

    print()
    print(title)
    print(debug.formatFuncCall(func, *args, **kwargs))
    print("---")
    term.table(
        [
            ["Multi", "Thread" if thread else "Process"],
            ["Number", number],
            ["Concurrency", concurrency],
            [],
            ["Cost Time:"],
            ["Total", "{}ms".format(roundns(totalTime))],
            ["Sum", "{}ms".format(roundns(sumTime))],
            ["Average", "{}ms".format(roundns(avgTime))],
            ["Min", "{}ms".format(roundns(minTime))],
            ["Max", "{}ms".format(roundns(maxTime))],
        ]
    )

    print()
    print("Percentage of the tests served within a certain time:")
    print("---")
    term.table(
        [
            [
                "{}%".format(int(mp * 100)),
                "{}ms".format(roundns(costTimes[int(number * mp) - 1])),
            ]
            for mp in markPercentages
        ]
    )

    return {
        "results": results,
        "cost": {
            "total": totalTime,
            "average": avgTime,
            "min": minTime,
            "max": maxTime,
        },
    }


def request(
    method,
    url,
    number,
    kwargs=None,
    concurrency=1,
    markPercentages=DEFAULT_MARK_PERCENTAGES,
):
    args = [method, url]
    result = test(
        requests.request,
        number=number,
        concurrency=concurrency,
        args=args,
        kwargs=kwargs,
        title="ABTest - Request - {}".format(method.upper()),
        markPercentages=markPercentages,
    )
    printResponseTime(result["results"], markPercentages=markPercentages)
    return result


def get(
    url,
    number,
    params=None,
    kwargs=None,
    concurrency=1,
    markPercentages=DEFAULT_MARK_PERCENTAGES,
):
    args = [url]
    kwargs = kwargs or {}
    kwargs.update(params=params)
    result = test(
        requests.get,
        number=number,
        concurrency=concurrency,
        args=args,
        kwargs=kwargs,
        title="ABTest - Request - GET",
        markPercentages=markPercentages,
    )
    printResponseTime(result["results"], markPercentages=markPercentages)
    return result


def post(
    url,
    number,
    data=None,
    json=None,
    kwargs=None,
    concurrency=1,
    markPercentages=DEFAULT_MARK_PERCENTAGES,
):
    args = [url]
    kwargs = kwargs or {}
    kwargs.update(data=data, json=json)
    result = test(
        requests.post,
        number=number,
        concurrency=concurrency,
        args=args,
        kwargs=kwargs,
        title="ABTest - Request - POST",
        markPercentages=markPercentages,
    )
    printResponseTime(result["results"], markPercentages=markPercentages)
    return result



# coding=utf-8
from __future__ import absolute_import, print_function

from contextlib import contextmanager

from mysql import connector

from suanpan.dw.rdbms import RelationalDataWarehouse


class MysqlDataWarehouse(RelationalDataWarehouse):
    @contextmanager
    def connect(self):
        connection = connector.connect(
            host=self.host,
            port=self.port,
            database=self.database,
            user=self.user,
            password=self.password,
        )

        yield connection
        connection.close()



# coding=utf-8
from __future__ import absolute_import, print_function

import contextlib
import itertools
from collections import defaultdict

from suanpan import runtime
from suanpan.arguments.auto import AutoArg
from suanpan.interfaces import HasArguments
from suanpan.log import logger
from suanpan.objects import Context


class Arguments(Context):
    pass


class Result(Context):
    pass


class Component(HasArguments):
    def __init__(self, funcOrComponent):
        if isinstance(funcOrComponent, Component):
            self.runFunc = funcOrComponent.runFunc
            self.arguments = funcOrComponent.arguments
        else:
            self.runFunc = funcOrComponent
            self.arguments = defaultdict(list)

    def __call__(self, *arg, **kwargs):
        self.run(*arg, **kwargs)

    @property
    def name(self):
        return self.runFunc.__name__

    @runtime.globalrun
    def run(self, *arg, **kwargs):
        self.beforeInit()
        context = self.init(*arg, **kwargs)
        self.afterInit()
        results = self.runFunc(context)
        self.beforeSave()
        outputs = self.save(context, results)
        self.afterSave()
        return outputs

    def init(self, *arg, **kwargs):
        restArgs = self.getArgList(*arg, **kwargs)
        globalArgs, restArgs = self.loadGlobalArguments(restArgs=restArgs)
        self.initBase(globalArgs)
        context = self._getContext(globalArgs)
        args, restArgs = self.loadComponentArguments(context, restArgs=restArgs)
        context.update(args=args)
        return context

    def beforeInit(self):
        pass

    def initBase(self, args):
        pass

    def afterInit(self):
        pass

    def save(self, context, results):
        outputs = self.saveOutputs(context, results)
        self._closeContext()
        return outputs

    def beforeSave(self):
        pass

    def afterSave(self):
        pass

    @contextlib.contextmanager
    def context(self, args=None):  # pylint: disable=unused-argument
        yield Context()

    def _getContext(self, *args, **kwargs):
        self.contextManager = self.context(  # pylint: disable=attribute-defined-outside-init
            *args, **kwargs
        )
        return next(self.contextManager.gen)  # pylint: disable-msg=e1101

    def _closeContext(self):
        try:
            next(self.contextManager.gen)  # pylint: disable-msg=e1101
        except StopIteration:
            pass

    def loadGlobalArguments(self, restArgs=None, **kwargs):
        logger.info("Loading Global Arguments:")
        arguments, restArgs = super(Component, self).loadGlobalArguments(
            restArgs=restArgs, **kwargs
        )
        args = self.argumentsDict(arguments)
        return Arguments.froms(args), restArgs

    def loadComponentArguments(self, context, restArgs=None):
        logger.info("Loading Component Arguments:")
        _, restArgs = self.loadFormatArguments(
            context, restArgs=restArgs, exclude="outputs"
        )
        _, restArgs = self.loadCleanArguments(
            context, restArgs=restArgs, include="outputs"
        )
        arguments = {
            k: self.argumentsDict(arg for arg in v if arg.isSet)
            for k, v in self.arguments.items()
        }
        arguments.update(self.argumentsDict(itertools.chain(*self.arguments.values())))
        return Arguments.froms(arguments), restArgs

    def getArguments(self, include=None, exclude=None):
        includes = set(self.arguments.keys() if not include else self._list(include))
        excludes = set([] if not exclude else self._list(exclude))
        includes = includes - excludes
        argumentsLists = [self.arguments[c] for c in includes]
        return list(itertools.chain(*argumentsLists))

    def saveMutipleOutputs(self, context, outputs, results):
        if isinstance(results, (tuple, list)):
            results = (Result.froms(value=result) for result in results)
            return self.saveArguments(context, outputs, results)
        if isinstance(results, dict):
            outputs, results = zip(
                *[
                    (
                        argument,
                        Result.froms(
                            value=self.getArgumentValueFromDict(results, argument)
                        ),
                    )
                    for argument in outputs
                    if self.getArgumentValueFromDict(results, argument) is not None
                ]
            )
            return self.saveArguments(context, outputs, results)
        raise Exception("Incorrect results: {}".format(results))

    def saveOneOutput(self, context, output, results):
        result = (
            Result.froms(value=self.getArgumentValueFromDict(results, output))
            if isinstance(results, dict)
            and self.hasArgumentValueFromDict(results, output)
            else Result.froms(value=results)
        )
        return {output.key: output.save(context, result)}

    def saveOutputs(self, context, results):
        logger.info("Saving...")
        outputs = self.getArguments(include="outputs")
        if len(outputs) > 1:
            return self.saveMutipleOutputs(context, outputs, results)
        if len(outputs) == 1:
            return self.saveOneOutput(context, outputs[0], results)
        return None

    def addArgument(self, arg, argtype="args", reverse=True):
        if reverse:
            self.arguments[argtype].insert(0, arg)
        else:
            self.arguments[argtype].append(arg)

    @classmethod
    def use(cls, funcOrComponent):
        return (
            funcOrComponent
            if isinstance(funcOrComponent, cls)
            else cls(funcOrComponent)
        )

    @classmethod
    def arg(cls, argument, argtype="args", reverse=True):
        def _dec(funcOrComponent):
            funcOrComponent = (
                funcOrComponent
                if isinstance(funcOrComponent, cls)
                else cls(funcOrComponent)
            )
            if isinstance(argument, AutoArg):
                argument.setBackend(argtype=argtype)
            funcOrComponent.addArgument(argument, argtype=argtype, reverse=reverse)
            return funcOrComponent

        return _dec

    @classmethod
    def input(cls, *args, **kwargs):
        kwargs.update(argtype="inputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def output(cls, *args, **kwargs):
        kwargs.update(argtype="outputs")
        return cls.arg(*args, **kwargs)

    @classmethod
    def param(cls, *args, **kwargs):
        kwargs.update(argtype="params")
        return cls.arg(*args, **kwargs)

    @classmethod
    def column(cls, *args, **kwargs):
        kwargs.update(argtype="columns")
        return cls.arg(*args, **kwargs)

    def _list(self, params=None):
        return [params] if isinstance(params, str) else list(params)



# coding=utf-8
from __future__ import absolute_import, print_function

import time
import traceback

import redis

from suanpan import utils
from suanpan.log import logger

MQ_DELIVER_KEY = "_suanpan_mq_deliver"


class RedisMQ(object):
    def __init__(
        self,  # pylint: disable=unused-argument
        redisHost="localhost",
        redisPort=6379,
        redisRealtime=False,
        redisKeepalive=True,
        # redisKeepaliveIDLE=120,
        # redisKeepaliveCNT=2,
        # redisKeepaliveINTVL=30,
        redisSocketConnectTimeout=1,
        options=None,
        client=None,
        **kwargs
    ):
        self.options = options or {}
        self.options.update(
            host=redisHost,
            port=redisPort,
            realtime=redisRealtime,
            keepalive=redisKeepalive,
            # keepidle=redisKeepaliveIDLE,
            # keepcnt=redisKeepaliveCNT,
            # keepintvl=redisKeepaliveINTVL,
            socketConnectTimeout=redisSocketConnectTimeout,
        )
        if client and not isinstance(client, redis.Redis):
            raise Exception("Invalid client: {}".format(client))
        self.client = client or redis.Redis(
            host=self.options["host"],
            port=self.options["port"],
            decode_responses=True,
            socket_keepalive=self.options["keepalive"],
            # socket_keepalive_options={
            #     socket.TCP_KEEPIDLE: self.options["keepidle"],
            #     socket.TCP_KEEPCNT: self.options["keepcnt"],
            #     socket.TCP_KEEPINTVL: self.options["keepintvl"],
            # },
            socket_connect_timeout=self.options["socketConnectTimeout"],
        )

    @property
    def connected(self):
        return bool(self.client.connection)

    def _getQueue(self, name):
        try:
            return self.client.xinfo_stream(name)
        except Exception:
            return {}

    def _getQueueGroups(self, name):
        try:
            return self.client.xinfo_groups(name)
        except Exception:
            return []

    def _lenQueue(self, name):
        try:
            return self.client.xlen(name)
        except Exception:
            return 0

    def listQueueNames(self):
        return self.client.keys("mq-*")

    def getQueueInfo(self, name):
        # queue = self._getQueue(name)
        length = self._lenQueue(name)
        groups = self._getQueueGroups(name)
        pending = sum(group.get("pending", 0) for group in groups)
        return {
            "name": name,
            "groups": groups,
            # "length": queue.get("length", 0),
            "length": length,
            "pending": pending,
        }

    def createQueue(self, name, group="default", consumeID="0", force=False, existOk=False):
        logger.debug("Create Queue {}-{}".format(name, group))
        if force:
            self.deleteQueue(name)
        return self._createQueue(name, group=group, consumeID=consumeID, existOk=existOk)

    def _createQueue(self, name, group="default", consumeID="0", existOk=False):
        try:
            return self.client.xgroup_create(name, group, id=consumeID, mkstream=True)
        except Exception:
            traceback.print_exc()
            if not existOk:
                raise Exception("Queue existed: {}".format(name))

    def deleteQueue(self, *names):
        return self.client.delete(*names)

    # def hasQueue(self, name, group="default"):
    #     try:
    #         queue = self.client.xinfo_stream(name)
    #         groups = self.client.xinfo_groups(name)
    #         return any(g["name"].decode() == group for g in groups)
    #     except Exception:
    #         return False

    def sendMessage(
        self, queue, data, messageID="*", maxlen=1000, trimImmediately=False
    ):
        return self.client.xadd(
            queue, data, id=messageID, maxlen=maxlen, approximate=(not trimImmediately)
        )

    def recvMessages(
        self,
        queue,
        group="default",
        consumer="unknown",
        noack=False,
        block=False,
        count=1,
        consumeID=">",
    ):
        block = None if not block else 0 if block is True else block
        messages = self.client.xreadgroup(
            group, consumer, {queue: consumeID}, count=count, block=block, noack=noack
        )
        messages = list(self._parseMessagesGenerator(messages, group))

        lostMessageIDs = [
            message["id"]
            for message in messages
            if message["id"] and not message["data"]
        ]
        if lostMessageIDs:
            self.client.xack(queue, group, *lostMessageIDs)
            logger.warning("Messages have lost: {}".format(lostMessageIDs))

        return [message for message in messages if message["data"]]

    def subscribeQueue(
        self,
        queue,
        group="default",
        consumer="unknown",
        noack=False,
        block=True,
        count=1,
        consumeID=">",
        delay=0,
        errDelay=1,
        errCallback=logger.error,
    ):
        self.createQueue(queue, group=group, existOk=True)
        logger.debug("Subscribing Messages")
        while True:
            try:
                messages = self.recvMessages(
                    queue,
                    group=group,
                    consumer=consumer,
                    noack=noack,
                    block=block,
                    count=count,
                    consumeID=consumeID,
                )
            except Exception as e:
                errCallback(e)
                logger.debug("Error in receiving messages. Wait {}s".format(errDelay))
                time.sleep(errDelay)
                self.createQueue(queue, group=group, existOk=True)
                continue

            if not messages:
                logger.debug("Received no messages. Wait {}s".format(delay))
                time.sleep(delay)
                continue

            for message in messages:
                yield message
                self.client.xack(queue, group, message["id"])

    def recvPendingMessagesInfo(
        self, queue, group="default", consumer="unknown", start="-", end="+", count=None
    ):
        return self.client.xpending_range(
            queue, group, start, end, count, consumername=consumer
        )

    def retryPendingMessages(
        self,
        queue,
        group="default",
        consumer="unknown",
        count=100,
        maxTimes=3,
        timeout=1,
        errCallback=logger.error,
        maxlen=1000,
        trimImmediately=False,
    ):
        logger.debug("Retrying Pending Messages")

        def _getPendingMessages():
            try:
                return self.recvMessages(
                    queue,
                    group=group,
                    consumer=consumer,
                    block=False,
                    count=count,
                    consumeID="0",
                )
            except Exception:
                logger.warning("Error in getting pending messages:")
                logger.warning(traceback.format_exc())
                return []

        pendingMessages = {msg["id"]: msg for msg in _getPendingMessages()}
        if not pendingMessages:
            logger.debug("Nothing to retry!")
            return

        pendingInfos = {
            msg["message_id"]: msg
            for msg in self.recvPendingMessagesInfo(
                queue, group=group, consumer=consumer, count=count
            )
        }

        for mid in pendingMessages.keys():
            message = pendingMessages[mid]
            info = pendingInfos.get(mid, {})
            message = utils.merge(message, info)
            data = message["data"]
            deliveredTimes = int(data.get(MQ_DELIVER_KEY, 1))
            if deliveredTimes >= maxTimes:
                logger.error(
                    "Message {} retry failed {} times. Drop!".format(
                        mid, deliveredTimes
                    )
                )
                errCallback(message)
                self.client.xack(queue, group, message["id"])
                continue
            timeSinceDelivered = message.get("time_since_delivered", 0)
            if timeSinceDelivered < timeout:
                logger.warning(
                    "Message {} maybe lost: {} < {}".format(
                        mid, timeSinceDelivered, timeout
                    )
                )
                continue
            success = self.client.xack(queue, group, message["id"])
            if success:
                data.update({MQ_DELIVER_KEY: deliveredTimes + 1})
                newMID = self.sendMessage(
                    queue, data, maxlen=maxlen, trimImmediately=trimImmediately
                )
                logger.warning(
                    "Message send back to queue: {} -> {}".format(mid, newMID)
                )

    def _parseMessagesGenerator(self, messages, group):
        for message in messages:
            queue, items = message
            for item in items:
                mid, data = item
                yield {"id": mid, "data": data, "queue": queue, "group": group}



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import Arg
from suanpan.mstorage import mstorage
from suanpan.utils import csv, json, npy, pickle


class MStorageArg(Arg):
    def __init__(self, *args, **kwargs):
        self.expire = kwargs.pop("expire", None)
        super(MStorageArg, self).__init__(*args, **kwargs)

    @property
    def isSet(self):
        return True

    def format(self, context):
        if self.value:
            self.value = self.getValue()
        return self.value

    def save(self, context, result):
        obj = result.value
        self.setValue(obj)
        self.logSaved(self.value)
        return self.value

    def getOutputTmpValue(self, *args):
        return "_".join(args)

    def getValue(self):
        raise NotImplementedError("Method not implemented!")

    def setValue(self, obj):
        raise NotImplementedError("Method not implemented!")


class Pickle(MStorageArg):
    def getValue(self):
        data = mstorage.mget(self.value)
        return pickle.loads(data)

    def setValue(self, obj):
        data = pickle.dumps(obj)
        return mstorage.set(self.value, data, expire=self.expire)


class Any(Pickle):
    pass


class Npy(MStorageArg):
    def getValue(self):
        data = mstorage.mget(self.value)
        params = json.loads(data["md"].decode())
        params["data"] = data["data"]
        return npy.loadjson(params)

    def setValue(self, obj):
        params = npy.dumpjson(obj)
        npybytes = params.pop("data")
        data = {"md": json.dumps(params), "data": npybytes}
        return mstorage.mset(self.value, data, expire=self.expire)


class Csv(Npy):
    def getValue(self):
        data = mstorage.get(self.value)
        return csv.loads(data)

    def setValue(self, dataframe):
        data = csv.dumps(dataframe)
        return mstorage.set(self.value, data, expire=self.expire)


class Table(Csv):
    pass
